// Linked list based queue after Sedgewick and Wayne, Algorithms 4th ed, Algorithm 1.3

#ifndef QUEUE_HPP
#define QUEUE_HPP

#include <cstddef>
#include <cassert>
#include <iterator>
#include <type_traits>

template <typename T>
class Queue {

private:

    ////////////////////////////////////
    class Node {
    public:
        T item;
        Node *next;
        };

public:

    ////////////////////////////////////

    template <bool is_const = true>
    class generic_iterator : public std::iterator<std::forward_iterator_tag, T> {

    private:
        typedef typename std::conditional<is_const, const Node *, Node *>::type NodePtrType;
        typedef typename std::conditional<is_const, const T&, T&>::type ValueReferenceType;

    public:

        generic_iterator( NodePtrType c ) :
            curr (c)
            {
            }

        generic_iterator( const generic_iterator<false>& other ) :
            curr (other.curr)
            {
            }

        generic_iterator & operator++() {
            if ( this->curr != nullptr ) {
                this->curr = this->curr->next;
                }
            return *this;
            }

        generic_iterator operator++( int ) {
            auto t = generic_iterator( *this );

            if ( this->curr != nullptr ) {
                this->curr = this->curr->next;
                }
            return t;
            }

        ValueReferenceType operator*() {
            return this->curr->item;
            }

        bool operator==( generic_iterator rhs ) {
            return this->curr == rhs.curr;
            }

        bool operator!=( generic_iterator rhs ) {
            return this->curr != rhs.curr;
            }

    private:

        NodePtrType curr;
        };

    ////////////////////////////////////

    typedef generic_iterator<true> const_iterator;
    typedef generic_iterator<false> iterator;
    typedef std::iterator_traits< generic_iterator<> > iterator_traits;

    Queue( void );
    ~Queue( void );

    // 1.3.41 Copy constructor.
    Queue( const Queue<T> &q );

    void enqueue( T &item );
    T dequeue( void );

    bool is_empty( void );
    size_t size( void );

    iterator begin( void );
    iterator end( void );
    const_iterator begin( void ) const;
    const_iterator end( void ) const;


private:

    size_t N;
    Node *first;
    Node *last;

    };


template <typename T>
Queue<T>::Queue( void ) :
    N (0),
    first (nullptr),
    last (nullptr)
    {
    }

// 1.3.41
template <typename T>
Queue<T>::Queue( const Queue<T> &q ) :
    N (0),
    first (nullptr),
    last (nullptr)
    {
    for ( auto it = q.begin(); it != q.end(); ++it ) {
        T item = *it;    
        this->enqueue( item );
        }
    }
    
template <typename T>
Queue<T>::~Queue( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    while ( first ) {
        Node *next = first->next;
        delete first;
        first = next;
        }
    }

template <typename T>
void Queue<T>::enqueue( T &item ) {
    Node * node = new Node();
    node->item = item;
    node->next = nullptr;

    if ( last ) {
        last->next = node;
        last = node;
        }
    else {
        last = first = node;
        }

    N++;
    }

template <typename T>
T Queue<T>::dequeue( void ) {
    assert( N > 0 );

    T item = first->item;

    Node *old = first;
    first = first->next;
    delete old;
    N--;

    if ( ! first )
        last = nullptr;

    return item;
    }

template <typename T>
bool Queue<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t Queue<T>::size( void ) {
    return N;
    }

template <typename T>
typename Queue<T>::iterator Queue<T>::begin( void ) {
    return Queue<T>::iterator( first );
    }

template <typename T>
typename Queue<T>::iterator Queue<T>::end( void ) {
    return Queue<T>::iterator( nullptr );
    }

template <typename T>
typename Queue<T>::const_iterator Queue<T>::begin( void ) const {
    return Queue<T>::const_iterator( first );
    }

template <typename T>
typename Queue<T>::const_iterator Queue<T>::end( void ) const {
    return Queue<T>::const_iterator( nullptr );
    }

#endif

