#ifndef RANDOMQUEUE_HPP
#define RANDOMQUEUE_HPP

#include <algorithm>
#include <cassert>
#include <cstddef>
#include <chrono>
#include <random>
#include <vector>

template <typename T>
class RandomQueue {

public:

    ////////////////////////////////////////
    class iterator {

    public:

        iterator( RandomQueue<T> *q, size_t c, size_t e );

        iterator& operator++();
        iterator operator++(int);

        T operator*();
        bool operator!=( typename RandomQueue<T>::iterator );

    private:

        RandomQueue<T> *queue;
        size_t curr;
        size_t end;
        };

    ////////////////////////////////////////
    RandomQueue( void );
    ~RandomQueue( void );

    void enqueue( T &item );
    T dequeue( void );
    T sample( void );

    bool is_empty( void );
    size_t size( void );

    iterator begin( void );
    iterator end( void );

private:

    size_t N;
    size_t max;
    T *data;
    std::vector<size_t> order;

    void resize( size_t new_max );
    size_t sample_index( void );
    void get_order( void );
    };


////////////////////////////////////////////////////////////////////////////////

template <typename T>
RandomQueue<T>::RandomQueue( void ) :
    N (0),
    max(1) {
    data = new T[2];
    }
    
template <typename T>
RandomQueue<T>::~RandomQueue( void ) {
    assert( N == 0 );  // Catch the potential memory leak.
    if ( data )
        delete[] data;
    }

template <typename T>
void RandomQueue<T>::enqueue( T &item ) {
    if ( N == max )
        resize( 2 * max );

    data[N] = item;
    ++N;
    }

template <typename T>
size_t RandomQueue<T>::sample_index( void ) {
    unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
    std::default_random_engine generator ( seed );
    std::uniform_int_distribution<size_t> distribution( 0, N-1 );
    return  distribution(generator);
    }

template <typename T>
T RandomQueue<T>::dequeue( void ) {
    assert( N > 0 );

    size_t i = sample_index();

    T tmp = data[N-1];
    data[N-1] = data[i];
    data[i] = tmp;

    T item = data[N-1];

    --N;

    return item;
    }

template <typename T>
T RandomQueue<T>::sample( void ) {
    assert( N > 0 );

    size_t i = sample_index();
    return data[i];
    }

template <typename T>
bool RandomQueue<T>::is_empty( void ) {
    return N == 0;
    }

template <typename T>
size_t RandomQueue<T>::size( void ) {
    return N;
    }

template <typename T>
void RandomQueue<T>::resize( size_t new_max ) {
    T *new_data = new T[new_max];
    for ( size_t i = 0; i < N; ++i )
        new_data[i] = data[i];
    T *old_data = data;
    data = new_data;
    delete[] old_data;
    max = new_max;
    }

template <typename T>
void RandomQueue<T>::get_order( void ) {
    for ( size_t i = 0; i < N; ++i )
        order.push_back( i );

    unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
    shuffle (order.begin(), order.end(), std::default_random_engine(seed));
    }

template <typename T>
typename RandomQueue<T>::iterator RandomQueue<T>::begin( void ) {
    if ( order.size() != N )
        get_order();

    return RandomQueue<T>::iterator( this, 0, order.size() );
    }

template <typename T>
typename RandomQueue<T>::iterator RandomQueue<T>::end( void ) {
    if ( order.size() != N )
        get_order();

    return RandomQueue<T>::iterator( this, order.size(), order.size() );
    }


////////////////////////////////////////////////////////////////////////////////

template <typename T>
RandomQueue<T>::iterator::iterator( RandomQueue<T> *q, size_t c, size_t e ) :
    queue (q),
    curr (c),
    end (e)
    {
    }

template <typename T>
typename RandomQueue<T>::iterator & RandomQueue<T>::iterator::operator++() {
    if ( this->curr != this->end )
        ++(this->curr);
    return *this;
    }

template <typename T>
typename RandomQueue<T>::iterator RandomQueue<T>::iterator::operator++( int ) {
    auto t = RandomQueue<T>::iterator( *this );

    if ( this->curr != this->end )
        ++(this->curr);

    return t;
    }

template <typename T>
T RandomQueue<T>::iterator::operator*() {

    return this->queue->data[ this->queue->order[ this->curr ] ];
    }

template <typename T>
bool RandomQueue<T>::iterator::operator!=( typename RandomQueue<T>::iterator other ) {
    return this->curr != other.curr;
    }


#endif

